;% Copyright (c) 1990-1994 The MITRE Corporation
;% 
;% Authors: W. M. Farmer, J. D. Guttman, F. J. Thayer
;%   
;% The MITRE Corporation (MITRE) provides this software to you without
;% charge to use, copy, modify or enhance for any legitimate purpose
;% provided you reproduce MITRE's copyright notice in any copy or
;% derivative work of this software.
;% 
;% This software is the copyright work of MITRE.  No ownership or other
;% proprietary interest in this software is granted you other than what
;% is granted in this license.
;% 
;% Any modification or enhancement of this software must identify the
;% part of this software that was modified, by whom and when, and must
;% inherit this license including its warranty disclaimers.
;% 
;% MITRE IS PROVIDING THE PRODUCT "AS IS" AND MAKES NO WARRANTY, EXPRESS
;% OR IMPLIED, AS TO THE ACCURACY, CAPABILITY, EFFICIENCY OR FUNCTIONING
;% OF THIS SOFTWARE AND DOCUMENTATION.  IN NO EVENT WILL MITRE BE LIABLE
;% FOR ANY GENERAL, CONSEQUENTIAL, INDIRECT, INCIDENTAL, EXEMPLARY OR
;% SPECIAL DAMAGES, EVEN IF MITRE HAS BEEN ADVISED OF THE POSSIBILITY OF
;% SUCH DAMAGES.
;% 
;% You, at your expense, hereby indemnify and hold harmless MITRE, its
;% Board of Trustees, officers, agents and employees, from any and all
;% liability or damages to third parties, including attorneys' fees,
;% court costs, and other related costs and expenses, arising out of your
;% use of this software irrespective of the cause of said liability.
;% 
;% The export from the United States or the subsequent reexport of this
;% software is subject to compliance with United States export control
;% and munitions control restrictions.  You agree that in the event you
;% seek to export this software or any derivative work thereof, you
;% assume full responsibility for obtaining all necessary export licenses
;% and approvals and for assuring compliance with applicable reexport
;% restrictions.
;% 
;% 
;% COPYRIGHT NOTICE INSERTED: Mon Apr 11 11:42:27 EDT 1994


(herald (resources two-d-table))

(let ((symbol-hash
       (*value t-implementation-env 'symbol-hash)
       ;;
       ;; (lambda (sym) (string-hash (symbol->string sym)))
       ;; 
       ))

(define-operation (two-d-table-hash obj)
  (cond ((fixnum? obj) obj)
	((symbol? obj) (symbol-hash obj))
	(else
	 (symbol-hash
	  (concatenate-symbol obj)))))

)

(define (make-two-d-table . maybe-id)
  (apply make-hash-table fixnum? identity eq? '#f maybe-id))

(define TWO-D-TABLE-ENTRY
  (let ((compute-hash
	 (lambda (key1 key2)
	   (imps-enforce
	    fixnum?
	    (hash-combine-two-fixnums
	     (imps-enforce fixnum? (two-d-table-hash key1))
	     (imps-enforce fixnum? (two-d-table-hash key2))))))
	(find-entry
	 (lambda (table hash key1 key2)
	   (let ((entry-list (table-entry table hash)))
	     (iterate iter ((entry-list entry-list))
	       (cond ((not entry-list) '#f)
		     ((let ((keys (caar entry-list)))
			(and (eq? key1 (car keys))
			     (eq? key2 (cdr keys))))
		      (car entry-list))
		     (else
		      (iter (cdr entry-list)))))))))
    (operation
	(lambda (table key1 key2)
	  (cond ((find-entry table (compute-hash key1 key2) key1 key2)
		 => cdr)
		(else '#f)))
      ((setter self)
       (lambda (table key1 key2 new-value)
	 (let ((hash (compute-hash key1 key2)))
	   (cond ((find-entry table hash key1 key2)
		  => (lambda (e) (set (cdr e) new-value)))
		 (else
		  (push (table-entry table hash)
			(cons (cons key1 key2) new-value))))))))))

(define (walk-two-d-table proc table)
  (walk-table
   (lambda (hash entry-list)
     (ignore hash)
     (walk
      (lambda (entry)
	(destructure ((((key1 . key2) . value) entry))
	  (proc key1 key2 value)))
      entry-list))
   table))
